@@ -14,9 +14,6 @@
 #    * limitations under the License.
 
 
-__author__ = 'dan'
-
-import datetime
 import json
 
 import pika
@@ -45,16 +42,15 @@ class AMQPClient(object):
         self.events_queue.queue_declare(queue=self.logs_queue_name, **settings)
 
     def publish_log(self, log):
-        self._publish(log, self.logs_queue_name, 'cloudify_log')
+        self._publish(log, self.logs_queue_name)
 
     def publish_event(self, event):
-        self._publish(event, self.events_queue_name, 'cloudify_event')
+        self._publish(event, self.events_queue_name)
 
-    def _publish(self, item, queue, message_type):
-        timestamp = str(datetime.datetime.now())
-        item['timestamp'] = timestamp
-        item['message_code'] = None
-        item['type'] = message_type
+    def close(self):
+        self.connection.close()
+
+    def _publish(self, item, queue):
         self.events_queue.basic_publish(exchange='',
                                         routing_key=queue,
                                         body=json.dumps(item))
