@@ -20,33 +20,35 @@ application root directory.
 """
 
 from __future__ import absolute_import
-from cloudify.constants import CELERY_WORK_DIR_PATH_KEY
 
 
-__author__ = 'idanmo'
-
 import sys
 import traceback
 import os
+from os import path
+
 from celery import Celery
 
+from cloudify.constants import CELERY_WORK_DIR_PATH_KEY
+
 
 celery = Celery('cloudify.celery',
-                broker='amqp://',
-                backend='amqp://')
+                broker=os.environ.get('BROKER_URL', 'amqp://'),
+                backend=os.environ.get('BROKER_URL', 'amqp://'))
 
-current_excepthook = sys.excepthook
+celery_work_folder = os.environ.get(CELERY_WORK_DIR_PATH_KEY)
 
+if celery_work_folder:
+    current_excepthook = sys.excepthook
 
-def new_excepthook(exception_type, value, the_traceback):
+    def new_excepthook(exception_type, value, the_traceback):
+        if not path.exists(celery_work_folder):
+            os.makedirs(celery_work_folder)
+        error_dump_path = path.join(celery_work_folder, 'celery_error.out')
+        with open(error_dump_path, 'w') as f:
+            f.write('Type: {0}\n'.format(exception_type))
+            f.write('Value: {0}\n'.format(value))
+            traceback.print_tb(the_traceback, file=f)
+        current_excepthook(exception_type, value, the_traceback)
 
-    work_folder = os.environ[CELERY_WORK_DIR_PATH_KEY]
-    if not os.path.exists(work_folder):
-        os.makedirs(work_folder)
-    with open(os.path.join(work_folder, 'celery_error.out'), 'w') as f:
-        f.write('Type: {0}\n'.format(exception_type))
-        f.write('Value: {0}\n'.format(value))
-        traceback.print_tb(the_traceback, file=f)
-    current_excepthook(type, value, the_traceback)
-
-sys.excepthook = new_excepthook
+    sys.excepthook = new_excepthook
