@@ -16,23 +16,23 @@
 
 import traceback
 import copy
-from multiprocessing import Process
-from multiprocessing import Pipe
+import sys
+import Queue
+from threading import Thread
 from StringIO import StringIO
 from functools import wraps
-import sys
 
 from cloudify import context
 from cloudify.workflows.workflow_context import CloudifyWorkflowContext
 from cloudify.manager import update_execution_status, get_rest_client
 from cloudify.workflows import api
 from cloudify_rest_client.executions import Execution
-from cloudify.exceptions import ProcessExecutionError
-from cloudify.exceptions import RecoverableError
-from cloudify.exceptions import NonRecoverableError
+from cloudify import exceptions
 from cloudify.state import current_ctx, current_workflow_ctx
 
-_stub_task = lambda fn: fn
+
+def _stub_task(fn):
+    return fn
 try:
     from cloudify.celery import celery as _celery
     _task = _celery.task
@@ -146,14 +146,28 @@ def operation(func=None, **arguments):
                 # preserve original type in the message
                 message = '{0}: {1}'.format(tpe.__name__, str(e))
 
-                if isinstance(e, NonRecoverableError):
-                    value = NonRecoverableError(message)
-                elif isinstance(e, RecoverableError):
-                    value = RecoverableError(message, e.retry_after)
+                # if the exception type is directly one of our exception
+                # than there is no need for conversion and we can just
+                # raise the original exception
+                if type(e) in [exceptions.OperationRetry,
+                               exceptions.RecoverableError,
+                               exceptions.NonRecoverableError,
+                               exceptions.HttpException]:
+                    raise
+
+                # if the exception inherits from our base exceptions, there
+                # still might be a de-serialization problem caused by one of
+                # the types in the inheritance tree.
+                if isinstance(e, exceptions.NonRecoverableError):
+                    value = exceptions.NonRecoverableError(message)
+                elif isinstance(e, exceptions.OperationRetry):
+                    value = exceptions.OperationRetry(message, e.retry_after)
+                elif isinstance(e, exceptions.RecoverableError):
+                    value = exceptions.RecoverableError(message, e.retry_after)
                 else:
                     # convert pure user exceptions
                     # to a RecoverableError
-                    value = RecoverableError(message)
+                    value = exceptions.RecoverableError(message)
 
                 raise type(value), value, tb
 
@@ -164,6 +178,8 @@ def operation(func=None, **arguments):
                 elif ctx.type == context.RELATIONSHIP_INSTANCE:
                     ctx.source.instance.update()
                     ctx.target.instance.update()
+            if ctx.operation._operation_retry:
+                raise ctx.operation._operation_retry
             return result
         return _process_wrapper(wrapper, arguments)
     else:
@@ -215,13 +231,17 @@ def workflow(func=None, **arguments):
         return partial_wrapper
 
 
+class RequestSystemExit(SystemExit):
+    pass
+
+
 def _remote_workflow(ctx, func, args, kwargs):
     def update_execution_cancelled():
         update_execution_status(ctx.execution_id, Execution.CANCELLED)
         _send_workflow_cancelled_event(ctx)
 
     rest = get_rest_client()
-    parent_conn, child_conn = Pipe()
+    parent_queue, child_queue = (Queue.Queue(), Queue.Queue())
     try:
         if rest.executions.get(ctx.execution_id).status in \
                 (Execution.CANCELLING, Execution.FORCE_CANCELLING):
@@ -234,17 +254,17 @@ def _remote_workflow(ctx, func, args, kwargs):
         _send_workflow_started_event(ctx)
 
         # the actual execution of the workflow will run in another
-        # process - this wrapper is the entry point for that
-        # process, and takes care of forwarding the result or error
-        # back to the parent process
+        # thread - this wrapper is the entry point for that
+        # thread, and takes care of forwarding the result or error
+        # back to the parent thread
         def child_wrapper():
             try:
                 ctx.internal.start_event_monitor()
                 workflow_result = _execute_workflow_function(
                     ctx, func, args, kwargs)
-                child_conn.send({'result': workflow_result})
+                child_queue.put({'result': workflow_result})
             except api.ExecutionCancelled:
-                child_conn.send({
+                child_queue.put({
                     'result': api.EXECUTION_CANCELLED_RESULT})
             except BaseException as workflow_ex:
                 tb = StringIO()
@@ -254,65 +274,71 @@ def _remote_workflow(ctx, func, args, kwargs):
                     'message': str(workflow_ex),
                     'traceback': tb.getvalue()
                 }
-                child_conn.send({'error': err})
+                child_queue.put({'error': err})
             finally:
-                child_conn.close()
+                ctx.internal.stop_event_monitor()
 
-        api.ctx = ctx
-        api.pipe = child_conn
+        api.queue = parent_queue
 
-        # starting workflow execution on child process
-        p = Process(target=child_wrapper)
-        p.start()
+        # starting workflow execution on child thread
+        t = Thread(target=child_wrapper)
+        t.start()
 
-        # while the child process is executing the workflow,
-        # the parent process is polling for 'cancel' requests while
-        # also waiting for messages from the child process
+        # while the child thread is executing the workflow,
+        # the parent thread is polling for 'cancel' requests while
+        # also waiting for messages from the child thread
         has_sent_cancelling_action = False
         result = None
+        execution = None
         while True:
-            # check if child process sent a message
-            if parent_conn.poll(5):
-                data = parent_conn.recv()
+            # check if child thread sent a message
+            try:
+                data = child_queue.get(timeout=5)
                 if 'result' in data:
-                    # child process has terminated
+                    # child thread has terminated
                     result = data['result']
                     break
                 else:
-                    # error occurred in child process
+                    # error occurred in child thread
                     error = data['error']
-                    raise ProcessExecutionError(error['message'],
-                                                error['type'],
-                                                error['traceback'])
+                    raise exceptions.ProcessExecutionError(error['message'],
+                                                           error['type'],
+                                                           error['traceback'])
+            except Queue.Empty:
+                pass
             # check for 'cancel' requests
             execution = rest.executions.get(ctx.execution_id)
             if execution.status == Execution.FORCE_CANCELLING:
-                # terminate the child process immediately
-                p.terminate()
                 result = api.EXECUTION_CANCELLED_RESULT
                 break
             elif not has_sent_cancelling_action and \
                     execution.status == Execution.CANCELLING:
-                # send a 'cancel' message to the child process. It
+                # send a 'cancel' message to the child thread. It
                 # is up to the workflow implementation to check for
                 # this message and act accordingly (by stopping and
-                # returning a api.EXECUTION_CANCELLED_RESULT result).
-                # parent process then goes back to polling for
+                # raising an api.ExecutionCancelled error, or by returning
+                # the deprecated api.EXECUTION_CANCELLED_RESULT as result).
+                # parent thread then goes back to polling for
                 # messages from child process or possibly
                 # 'force-cancelling' requests
-                parent_conn.send({'action': 'cancel'})
+                parent_queue.put({'action': 'cancel'})
                 has_sent_cancelling_action = True
 
         # updating execution status and sending events according to
         # how the execution ended
         if result == api.EXECUTION_CANCELLED_RESULT:
             update_execution_cancelled()
+            if execution and execution.status == Execution.FORCE_CANCELLING:
+                # TODO: kill worker externally
+                raise RequestSystemExit()
         else:
             update_execution_status(ctx.execution_id, Execution.TERMINATED)
             _send_workflow_succeeded_event(ctx)
         return result
+    except RequestSystemExit:
+        raise
     except BaseException as e:
-        if isinstance(e, ProcessExecutionError):
+        if isinstance(e, exceptions.ProcessExecutionError):
             error_traceback = e.traceback
         else:
             error = StringIO()
@@ -322,9 +348,6 @@ def _remote_workflow(ctx, func, args, kwargs):
                                 error_traceback)
         _send_workflow_failed_event(ctx, e, error_traceback)
         raise
-    finally:
-        parent_conn.close()
-        child_conn.close()  # probably unneeded but cleanup anyway
 
 
 def _local_workflow(ctx, func, args, kwargs):
