@@ -13,7 +13,6 @@
 #    * See the License for the specific language governing permissions and
 #    * limitations under the License.
 
-
 from cloudify.decorators import workflow
 from cloudify.workflows.tasks_graph import forkjoin
 from cloudify.workflows import tasks as workflow_tasks
@@ -119,26 +118,25 @@ class InstallationTasksGraphFinisher(object):
                 self._enforce_correct_src_trg_order(instance, rel)
 
 
-class AutohealInstallationTasksGraphFinisher(InstallationTasksGraphFinisher):
+class RuntimeInstallationTasksGraphFinisher(InstallationTasksGraphFinisher):
     def _enforce_correct_src_trg_order(self, instance, rel):
         # Handle only nodes within self.node_instances, others are running
         if rel.target_node_instance in self.node_instances:
-            super(AutohealInstallationTasksGraphFinisher,
+            super(RuntimeInstallationTasksGraphFinisher,
                   self)._enforce_correct_src_trg_order(instance, rel)
 
     def finish_creation(self):
-        super(AutohealInstallationTasksGraphFinisher, self).finish_creation()
+        super(RuntimeInstallationTasksGraphFinisher, self).finish_creation()
         # Add operations for intact nodes depending on a node instance
         # belonging to node_instances (which are being reinstalled)
         for instance in self.intact_nodes:
             for rel in instance.relationships:
                 if rel.target_node_instance in self.node_instances:
                     trg_started = self.tasks.set_state_started[rel.target_id]
-                    establish_operations = _relationship_operations(
-                        instance,
-                        'cloudify.interfaces.relationship_lifecycle.establish'
-                    )
-                    for establish_op in establish_operations:
+                    establish_ops = _relationship_operations_with_target(
+                        rel,
+                        'cloudify.interfaces.relationship_lifecycle.establish')
+                    for establish_op, _ in establish_ops:
                         self.graph.add_task(establish_op)
                         self.graph.add_dependency(establish_op, trg_started)
 
@@ -173,7 +171,7 @@ def _install_node_instances(ctx, node_instances, intact_nodes,
         tasks
     ).finish_creation()
 
-    return graph.execute()
+    graph.execute()
 
 
 class UninstallationTasksReferences(object):
@@ -259,24 +257,23 @@ class UninstallationTasksGraphFinisher(object):
                 self._enforce_correct_src_trg_order(instance, rel)
 
 
-class AutohealUninstallationTasksGraphFinisher(
+class RuntimeUninstallationTasksGraphFinisher(
         UninstallationTasksGraphFinisher):
 
     def _enforce_correct_src_trg_order(self, instance, rel):
         if rel.target_node_instance in self.node_instances:
-            super(AutohealUninstallationTasksGraphFinisher,
+            super(RuntimeUninstallationTasksGraphFinisher,
                   self)._enforce_correct_src_trg_order(instance, rel)
 
     def finish_creation(self):
-        super(AutohealUninstallationTasksGraphFinisher, self).finish_creation()
+        super(RuntimeUninstallationTasksGraphFinisher, self).finish_creation()
         for instance in self.intact_nodes:
             for rel in instance.relationships:
                 if rel.target_node_instance in self.node_instances:
                     target_stopped = self.tasks.stop_node[rel.target_id]
-                    unlink_tasks = _relationship_operations_with_targets(
-                        instance,
-                        'cloudify.interfaces.relationship_lifecycle.unlink'
-                    )
+                    unlink_tasks = _relationship_operations_with_target(
+                        rel,
+                        'cloudify.interfaces.relationship_lifecycle.unlink')
                     for unlink_task, _ in unlink_tasks:
                         self.graph.add_task(unlink_task)
                         self.graph.add_dependency(unlink_task, target_stopped)
@@ -285,10 +282,12 @@ class AutohealUninstallationTasksGraphFinisher(
 
 
 def _uninstall_node_instances(ctx, node_instances, intact_nodes,
-                              node_tasks_seq_creator, graph_finisher_cls):
-    # switch to graph mode (operations on the context return tasks instead of
-    # result instances)
-    graph = ctx.graph_mode()
+                              node_tasks_seq_creator, graph_finisher_cls,
+                              graph=None):
+    if not graph:
+        # switch to graph mode (operations on the context return tasks instead
+        # of result instances)
+        graph = ctx.graph_mode()
     tasks_refs = UninstallationTasksReferences()
     for instance in node_instances:
         # We need reference to the set deleted state tasks and the set
@@ -324,7 +323,7 @@ def _uninstall_node_instances(ctx, node_instances, intact_nodes,
         tasks_refs
     ).finish_creation()
 
-    return graph.execute()
+    graph.execute()
 
 
 def _set_send_node_event_on_error_handler(task, node_instance, error_message):
@@ -353,15 +352,19 @@ def _relationship_operations(node_instance, operation):
 def _relationship_operations_with_targets(node_instance, operation):
     tasks = []
     for relationship in node_instance.relationships:
-        tasks.append(
-            (relationship.execute_source_operation(operation),
-             relationship.target_id))
-        tasks.append(
-            (relationship.execute_target_operation(operation),
-             relationship.target_id))
+        tasks += _relationship_operations_with_target(relationship, operation)
     return tasks
 
 
+def _relationship_operations_with_target(relationship, operation):
+    return [
+        (relationship.execute_source_operation(operation),
+         relationship.target_id),
+        (relationship.execute_target_operation(operation),
+         relationship.target_id)
+    ]
+
+
 def _is_host_node(node_instance):
     return 'cloudify.nodes.Compute' in node_instance.node.type_hierarchy
 
@@ -446,7 +449,7 @@ def _host_pre_stop(host_node_instance):
 
 
 @workflow
-def execute_operation(ctx, operation, operation_kwargs,
+def execute_operation(ctx, operation, operation_kwargs, allow_kwargs_override,
                       run_by_dependency_order, type_names, node_ids,
                       node_instance_ids, **kwargs):
     """ A generic workflow for executing arbitrary operations on nodes """
@@ -506,12 +509,19 @@ def execute_operation(ctx, operation, operation_kwargs,
                     send_event_done_tasks[instance.id] = nop_task
                     graph.add_task(nop_task)
 
+    # preparing the parameters to the execute_operation call
+    exec_op_params = dict()
+    exec_op_params['kwargs'] = operation_kwargs
+    exec_op_params['operation'] = operation
+    if allow_kwargs_override is not None:
+        exec_op_params['allow_kwargs_override'] = allow_kwargs_override
+
     # registering actual tasks to sequences
     for instance in filtered_node_instances:
         sequence = graph.sequence()
         sequence.add(
             send_event_starting_tasks[instance.id],
-            instance.execute_operation(operation, kwargs=operation_kwargs),
+            instance.execute_operation(**exec_op_params),
             send_event_done_tasks[instance.id])
 
     # adding tasks dependencies if required
@@ -527,7 +537,7 @@ def execute_operation(ctx, operation, operation_kwargs,
                     graph.add_dependency(instance_starting_task,
                                          target_done_task)
 
-    return graph.execute()
+    graph.execute()
 
 
 @workflow
@@ -559,8 +569,8 @@ def uninstall(ctx, **kwargs):
 @workflow
 def auto_heal_reinstall_node_subgraph(
         ctx,
-        node_id,
-        diagnose_value=None,
+        node_instance_id,
+        diagnose_value='Not provided',
         **kwargs):
     """Reinstalls the whole subgraph of the system topology
 
@@ -573,10 +583,9 @@ def auto_heal_reinstall_node_subgraph(
     :param diagnose_value: diagnosed reason of failure
     """
 
-    failing_node = ctx.get_node_instance(node_id)
-    # TODO: this is a workaround until a better fix is introduced
-    if failing_node.get_state().get() != "started":
-        return
+    ctx.logger.info("Starting 'heal' workflow on {0}, Diagnosis: {1}"
+                    .format(node_instance_id, diagnose_value))
+    failing_node = ctx.get_node_instance(node_instance_id)
     failing_node_host = ctx.get_node_instance(
         failing_node._node_instance.host_id
     )
@@ -587,12 +596,141 @@ def auto_heal_reinstall_node_subgraph(
         subgraph_node_instances,
         intact_nodes,
         NodeUninstallationTasksSequenceCreator(),
-        AutohealUninstallationTasksGraphFinisher
+        RuntimeUninstallationTasksGraphFinisher
     )
     _install_node_instances(
         ctx,
         subgraph_node_instances,
         intact_nodes,
         NodeInstallationTasksSequenceCreator(),
-        AutohealInstallationTasksGraphFinisher
+        RuntimeInstallationTasksGraphFinisher
     )
+
+
+@workflow
+def scale(ctx, node_id, delta, scale_compute, **kwargs):
+    """Scales in/out the subgraph of node_id.
+
+    If `scale_compute` is set to false, the subgraph will consist of all
+    the nodes that are contained in `node_id` and `node_id` itself.
+    If `scale_compute` is set to true, the subgraph will consist of all
+    nodes that are contained in the compute node that contains `node_id`
+    and the compute node itself.
+    If `node_id` is not contained in a compute node and is not a compute node,
+    this property is ignored.
+
+    `delta` is used to specify the scale factor.
+    For `delta > 0`: If current number of instances is `N`, scale out to
+    `N + delta`.
+    For `delta < 0`: If current number of instances is `N`, scale in to
+    `N - |delta|`.
+
+    :param ctx: cloudify context
+    :param node_id: the node_id to scale
+    :param delta: scale in/out factor
+    :param scale_compute: should scale apply on compute node containing
+                          'node_id'
+    """
+
+    node = ctx.get_node(node_id)
+    if not node:
+        raise ValueError("Node {0} doesn't exist".format(node_id))
+    if delta == 0:
+        ctx.logger.info('delta parameter is 0, so no scaling will take place.')
+        return
+    host_node = node.host_node
+    scaled_node = host_node if (scale_compute and host_node) else node
+    curr_num_instances = scaled_node.number_of_instances
+    planned_num_instances = curr_num_instances + delta
+    if planned_num_instances < 1:
+        raise ValueError('Provided delta: {0} is illegal. current number of'
+                         'instances of node {1} is {2}'
+                         .format(delta, node_id, curr_num_instances))
+
+    modification = ctx.deployment.start_modification({
+        scaled_node.id: {
+            'instances': planned_num_instances
+
+            # These following parameters are not exposed at the moment,
+            # but should be used to control which node instances get scaled in
+            # (when scaling in).
+            # They are mentioned here, because currently, the modification API
+            # is not very documented.
+            # Special care should be taken because if `scale_compute == True`
+            # (which is the default), then these ids should be the compute node
+            # instance ids which are not necessarily instances of the node
+            # specified by `node_id`.
+
+            # Node instances denoted by these instance ids should be *kept* if
+            # possible.
+            # 'removed_ids_exclude_hint': [],
+
+            # Node instances denoted by these instance ids should be *removed*
+            # if possible.
+            # 'removed_ids_include_hint': []
+        }
+    })
+    try:
+        ctx.logger.info('Deployment modification started. '
+                        '[modification_id={0}]'.format(modification.id))
+        if delta > 0:
+            added_and_related = _get_all_nodes_instances(modification.added)
+            added = set(i for i in added_and_related
+                        if i.modification == 'added')
+            related = added_and_related - added
+            try:
+                node_tasks_seq = NodeInstallationTasksSequenceCreator()
+                _install_node_instances(
+                    ctx,
+                    node_instances=added,
+                    intact_nodes=related,
+                    node_tasks_seq_creator=node_tasks_seq,
+                    graph_finisher_cls=RuntimeInstallationTasksGraphFinisher)
+            except:
+                ctx.logger.error('Scale out failed, scaling back in.')
+                graph = ctx.internal.task_graph
+                for task in graph.tasks_iter():
+                    graph.remove_task(task)
+                node_tasks_seq = NodeUninstallationTasksSequenceCreator()
+                _uninstall_node_instances(
+                    ctx,
+                    node_instances=added,
+                    intact_nodes=related,
+                    node_tasks_seq_creator=node_tasks_seq,
+                    graph_finisher_cls=RuntimeUninstallationTasksGraphFinisher,
+                    graph=graph)
+                raise
+        else:
+            removed_and_related = _get_all_nodes_instances(
+                modification.removed)
+            removed = set(i for i in removed_and_related
+                          if i.modification == 'removed')
+            related = removed_and_related - removed
+            node_tasks_seq = NodeUninstallationTasksSequenceCreator()
+            _uninstall_node_instances(
+                ctx,
+                node_instances=removed,
+                intact_nodes=related,
+                node_tasks_seq_creator=node_tasks_seq,
+                graph_finisher_cls=RuntimeUninstallationTasksGraphFinisher)
+    except:
+        ctx.logger.warn('Rolling back deployment modification. '
+                        '[modification_id={0}]'.format(modification.id))
+        try:
+            modification.rollback()
+        except:
+            ctx.logger.warn('Deployment modification rollback failed. The '
+                            'deployment model is most likely in some corrupted'
+                            ' state.'
+                            '[modification_id={0}]'.format(modification.id))
+            raise
+        raise
+    else:
+        try:
+            modification.finish()
+        except:
+            ctx.logger.warn('Deployment modification finish failed. The '
+                            'deployment model is most likely in some corrupted'
+                            ' state.'
+                            '[modification_id={0}]'.format(modification.id))
+            raise
