@@ -13,14 +13,15 @@
 #    * See the License for the specific language governing permissions and
 #    * limitations under the License.
 
-
 import os
 import tempfile
 import copy
 import importlib
+import shutil
 import uuid
 import json
 import threading
+from contextlib import contextmanager
 
 from cloudify_rest_client.nodes import Node
 from cloudify_rest_client.node_instances import NodeInstance
@@ -29,12 +30,14 @@ from cloudify.workflows.workflow_context import (
     DEFAULT_LOCAL_TASK_THREAD_POOL_SIZE)
 
 try:
+    from dsl_parser.parser import HOST_TYPE
     from dsl_parser import parser as dsl_parser, tasks as dsl_tasks
     from dsl_parser import functions as dsl_functions
 except ImportError:
     dsl_parser = None
     dsl_tasks = None
     dsl_functions = None
+    HOST_TYPE = None
 
 
 class _Environment(object):
@@ -60,8 +63,7 @@ class _Environment(object):
                 plan=plan,
                 nodes=nodes,
                 node_instances=node_instances,
-                resources_root=os.path.dirname(
-                    os.path.abspath(blueprint_path)))
+                blueprint_path=blueprint_path)
 
     @property
     def plan(self):
@@ -159,6 +161,18 @@ def _parse_plan(blueprint_path, inputs, ignored_modules):
     return plan, nodes, node_instances
 
 
+def _validate_node(node):
+    if HOST_TYPE in node['type_hierarchy']:
+        install_agent_prop = node.properties.get('install_agent')
+        if install_agent_prop:
+            raise ValueError("'install_agent': true is not supported "
+                             "(it is True by default) "
+                             "when executing local workflows. "
+                             "The 'install_agent' property "
+                             "must be set to false for each node of type {0}."
+                             .format(HOST_TYPE))
+
+
 def _prepare_nodes_and_instances(nodes, node_instances, ignored_modules):
 
     def scan(parent, name, node):
@@ -171,9 +185,13 @@ def _prepare_nodes_and_instances(nodes, node_instances, ignored_modules):
                                ignored_modules=ignored_modules)
 
     for node in nodes:
+        number_of_instances = node['instances']['deploy']
+        node['number_of_instances'] = number_of_instances
+        node['deploy_number_of_instances'] = number_of_instances
         if 'relationships' not in node:
             node['relationships'] = []
         scan(node, 'operations', node)
+        _validate_node(node)
         for relationship in node['relationships']:
             scan(relationship, 'source_operations', node)
             scan(relationship, 'target_operations', node)
@@ -260,9 +278,9 @@ class _Storage(object):
         self._locks = None
         self.env = None
 
-    def init(self, name, plan, nodes, node_instances, resources_root):
+    def init(self, name, plan, nodes, node_instances, blueprint_path):
         self.name = name
-        self.resources_root = resources_root
+        self.resources_root = os.path.dirname(os.path.abspath(blueprint_path))
         self.plan = plan
         self._init_locks_and_nodes(nodes)
 
@@ -283,7 +301,7 @@ class _Storage(object):
             suffix = '-{0}'.format(os.path.basename(resource_path))
             target_path = tempfile.mktemp(suffix=suffix)
         resource = self.get_resource(resource_path)
-        with open(target_path, 'w') as f:
+        with open(target_path, 'wb') as f:
             f.write(resource)
         return target_path
 
@@ -351,12 +369,12 @@ class InMemoryStorage(_Storage):
         super(InMemoryStorage, self).__init__()
         self._node_instances = None
 
-    def init(self, name, plan, nodes, node_instances, resources_root):
+    def init(self, name, plan, nodes, node_instances, blueprint_path):
         self.plan = plan
         self._node_instances = dict((instance.id, instance)
                                     for instance in node_instances)
         super(InMemoryStorage, self).init(name, plan, nodes, node_instances,
-                                          resources_root)
+                                          blueprint_path)
 
     def load(self, name):
         raise NotImplementedError('load is not implemented by memory storage')
@@ -385,19 +403,33 @@ class FileStorage(_Storage):
         self._storage_dir = None
         self._instances_dir = None
         self._data_path = None
+        self._payload_path = None
+        self._blueprint_path = None
 
-    def init(self, name, plan, nodes, node_instances, resources_root):
+    def init(self, name, plan, nodes, node_instances, blueprint_path):
         storage_dir = os.path.join(self._root_storage_dir, name)
         instances_dir = os.path.join(storage_dir, 'node-instances')
         data_path = os.path.join(storage_dir, 'data')
+        payload_path = os.path.join(storage_dir, 'payload')
         os.makedirs(storage_dir)
         os.mkdir(instances_dir)
+        with open(payload_path, 'w') as f:
+            f.write(json.dumps({}))
+
+        blueprint_filename = os.path.basename(os.path.abspath(blueprint_path))
         with open(data_path, 'w') as f:
             f.write(json.dumps({
                 'plan': plan,
-                'resources_root': resources_root,
+                'blueprint_filename': blueprint_filename,
                 'nodes': nodes
             }))
+        resources_root = os.path.dirname(os.path.abspath(blueprint_path))
+        self.resources_root = os.path.join(storage_dir, 'resources')
+
+        def ignore(src, names):
+            return names if os.path.abspath(self.resources_root) == src \
+                else set()
+        shutil.copytree(resources_root, self.resources_root, ignore=ignore)
         self._instances_dir = instances_dir
         for instance in node_instances:
             self._store_instance(instance, lock=False)
@@ -407,14 +439,29 @@ class FileStorage(_Storage):
         self.name = name
         self._storage_dir = os.path.join(self._root_storage_dir, name)
         self._instances_dir = os.path.join(self._storage_dir, 'node-instances')
+        self._payload_path = os.path.join(self._storage_dir, 'payload')
         self._data_path = os.path.join(self._storage_dir, 'data')
         with open(self._data_path) as f:
             data = json.loads(f.read())
         self.plan = data['plan']
-        self.resources_root = data['resources_root']
+        self.resources_root = os.path.join(self._storage_dir, 'resources')
+        self._blueprint_path = os.path.join(self.resources_root,
+                                            data['blueprint_filename'])
         nodes = [Node(node) for node in data['nodes']]
         self._init_locks_and_nodes(nodes)
 
+    @contextmanager
+    def payload(self):
+        with open(self._payload_path, 'r') as f:
+            payload = json.load(f)
+            yield payload
+        with open(self._payload_path, 'w') as f:
+            json.dump(payload, f, indent=2)
+            f.write(os.linesep)
+
+    def get_blueprint_path(self):
+        return self._blueprint_path
+
     def get_node_instance(self, node_instance_id):
         return self._get_node_instance(node_instance_id)
 
@@ -424,6 +471,7 @@ class FileStorage(_Storage):
                 return NodeInstance(json.loads(f.read()))
 
     def _store_instance(self, node_instance, lock=True):
+        instance_lock = None
         if lock:
             instance_lock = self._lock(node_instance.id)
             instance_lock.acquire()
@@ -431,7 +479,7 @@ class FileStorage(_Storage):
             with open(self._instance_path(node_instance.id), 'w') as f:
                 f.write(json.dumps(node_instance))
         finally:
-            if lock:
+            if lock and instance_lock:
                 instance_lock.release()
 
     def _instance_path(self, node_instance_id):
