@@ -13,30 +13,31 @@
 #    * See the License for the specific language governing permissions and
 #    * limitations under the License.
 
-__author__ = 'dank'
 
 import sys
 import time
 import uuid
 import Queue
 
-from cloudify.celery import celery as celery_client
-from cloudify.exceptions import NonRecoverableError, RecoverableError
-
+from cloudify import exceptions
 from cloudify.workflows import api
 
 INFINITE_TOTAL_RETRIES = -1
 DEFAULT_TOTAL_RETRIES = INFINITE_TOTAL_RETRIES
 DEFAULT_RETRY_INTERVAL = 30
 
+DEFAULT_SEND_TASK_EVENTS = True
 
 TASK_PENDING = 'pending'
 TASK_SENDING = 'sending'
 TASK_SENT = 'sent'
 TASK_STARTED = 'started'
+TASK_RESCHEDULED = 'rescheduled'
 TASK_SUCCEEDED = 'succeeded'
 TASK_FAILED = 'failed'
 
+TERMINATED_STATES = [TASK_RESCHEDULED, TASK_SUCCEEDED, TASK_FAILED]
+
 
 def retry_failure_handler(task):
     """Basic on_success/on_failure handler that always returns retry"""
@@ -47,13 +48,14 @@ class WorkflowTask(object):
     """A base class for workflow tasks"""
 
     def __init__(self,
+                 workflow_context,
                  task_id=None,
                  info=None,
                  on_success=None,
                  on_failure=None,
                  total_retries=DEFAULT_TOTAL_RETRIES,
                  retry_interval=DEFAULT_RETRY_INTERVAL,
-                 workflow_context=None):
+                 send_task_events=DEFAULT_SEND_TASK_EVENTS):
         """
         :param task_id: The id of this task (generated if none is provided)
         :param info: A short description of this task (for logging)
@@ -86,7 +88,9 @@ class WorkflowTask(object):
         self.total_retries = total_retries
         self.retry_interval = retry_interval
         self.terminated = Queue.Queue(maxsize=1)
+        self.is_terminated = False
         self.workflow_context = workflow_context
+        self.send_task_events = send_task_events
 
         self.current_retries = 0
         # timestamp for which the task should not be executed
@@ -94,6 +98,16 @@ class WorkflowTask(object):
         # graph during retries
         self.execute_after = time.time()
 
+    def dump(self):
+        return {
+            'id': self.id,
+            'state': self.get_state(),
+            'info': self.info,
+            'error': self.error,
+            'current_retries': self.current_retries,
+            'cloudify_context': self.cloudify_context
+        }
+
     def is_remote(self):
         """
         :return: Is this a remote task
@@ -116,8 +130,8 @@ class WorkflowTask(object):
         """
         Get the task state
 
-        :return: The task state [pending, sending, sent, started, succeeded,
-                                 failed]
+        :return: The task state [pending, sending, sent, started,
+                                 rescheduled, succeeded, failed]
         """
         return self._state
 
@@ -126,24 +140,25 @@ class WorkflowTask(object):
         Set the task state
 
         :param state: The state to set [pending, sending, sent, started,
-                                           succeeded, failed]
+                                        rescheduled, succeeded, failed]
         """
-
         if state not in [TASK_PENDING, TASK_SENDING, TASK_SENT, TASK_STARTED,
-                         TASK_SUCCEEDED, TASK_FAILED]:
-            raise RuntimeError('Illegal state set on task: {} '
-                               '[task={}]'.format(state, str(self)))
-
+                         TASK_RESCHEDULED, TASK_SUCCEEDED, TASK_FAILED]:
+            raise RuntimeError('Illegal state set on task: {0} '
+                               '[task={1}]'.format(state, str(self)))
         self._state = state
-        if state in [TASK_SUCCEEDED, TASK_FAILED]:
+        if state in TERMINATED_STATES:
+            self.is_terminated = True
             self.terminated.put_nowait(True)
 
     def wait_for_terminated(self, timeout=None):
+        if self.is_terminated:
+            return
         self.terminated.get(timeout=timeout)
 
     def handle_task_terminated(self):
-        if self.get_state() == TASK_FAILED:
-            handler_result = self._handle_task_failed()
+        if self.get_state() in (TASK_FAILED, TASK_RESCHEDULED):
+            handler_result = self._handle_task_not_succeeded()
         else:
             handler_result = self._handle_task_succeeded()
 
@@ -153,10 +168,8 @@ class WorkflowTask(object):
                     handler_result.ignore_total_retries]):
                 if handler_result.retry_after is None:
                     handler_result.retry_after = self.retry_interval
-                new_task = self.duplicate()
-                new_task.current_retries += 1
-                new_task.execute_after = time.time() +\
-                    handler_result.retry_after
+                new_task = self.duplicate_for_retry(
+                    time.time() + handler_result.retry_after)
                 handler_result.retried_task = new_task
             else:
                 handler_result.action = HandlerResult.HANDLER_FAIL
@@ -169,32 +182,67 @@ class WorkflowTask(object):
         else:
             return HandlerResult.cont()
 
-    def _handle_task_failed(self):
-        """Call handler for task failure"""
-        if self.on_failure:
+    def _handle_task_not_succeeded(self):
+
+        """
+        Call handler for task which hasn't ended in 'succeeded' state
+        (i.e. has either failed or been rescheduled)
+        """
+
+        try:
+            exception = self.async_result.result
+        except Exception as e:
+            exception = exceptions.NonRecoverableError(
+                'Could not de-serialize '
+                'exception of task {0} --> {1}: {2}'
+                .format(self.name,
+                        type(e).__name__,
+                        str(e)))
+
+        if isinstance(exception, exceptions.OperationRetry):
+            # operation explicitly requested a retry, so we ignore
+            # the handler set on the task.
+            handler_result = HandlerResult.retry()
+        elif self.on_failure:
             handler_result = self.on_failure(self)
         else:
             handler_result = HandlerResult.retry()
-        if handler_result.action == HandlerResult.HANDLER_RETRY and \
-                self.is_remote():
-            try:
-                exception = self.async_result.async_result.result
-            except:
-                exception = None
-            if isinstance(exception, NonRecoverableError):
+
+        if handler_result.action == HandlerResult.HANDLER_RETRY:
+            if isinstance(exception, exceptions.NonRecoverableError):
                 handler_result = HandlerResult.fail()
-            elif isinstance(exception, RecoverableError):
+            elif isinstance(exception, exceptions.RecoverableError):
                 handler_result.retry_after = exception.retry_after
         return handler_result
 
     def __str__(self):
         suffix = self.info if self.info is not None else ''
-        return '{}({})'.format(self.name, suffix)
+        return '{0}({1})'.format(self.name, suffix)
 
-    def duplicate(self):
+    def duplicate_for_retry(self, execute_after):
         """
         :return: A new instance of this task with a new task id
         """
+        dup = self._duplicate()
+        dup.execute_after = execute_after
+        dup.current_retries = self.current_retries + 1
+        if dup.cloudify_context and 'operation' in dup.cloudify_context:
+            op_ctx = dup.cloudify_context['operation']
+            op_ctx['retry_number'] = dup.current_retries
+        return dup
+
+    def _duplicate(self):
+        raise NotImplementedError('Implemented by subclasses')
+
+    @property
+    def cloudify_context(self):
+        raise NotImplementedError('Implemented by subclasses')
+
+    @property
+    def name(self):
+        """
+        :return: The task name
+        """
 
         raise NotImplementedError('Implemented by subclasses')
 
@@ -208,13 +256,14 @@ class RemoteWorkflowTask(WorkflowTask):
     def __init__(self,
                  task,
                  cloudify_context,
+                 workflow_context,
                  task_id=None,
                  info=None,
                  on_success=None,
                  on_failure=retry_failure_handler,
                  total_retries=DEFAULT_TOTAL_RETRIES,
                  retry_interval=DEFAULT_RETRY_INTERVAL,
-                 workflow_context=None):
+                 send_task_events=DEFAULT_SEND_TASK_EVENTS):
         """
         :param task: The celery task
         :param cloudify_context: the cloudify context dict
@@ -240,15 +289,16 @@ class RemoteWorkflowTask(WorkflowTask):
         :param workflow_context: the CloudifyWorkflowContext instance
         """
         super(RemoteWorkflowTask, self).__init__(
+            workflow_context,
             task_id,
             info=info,
             on_success=on_success,
             on_failure=on_failure,
             total_retries=total_retries,
             retry_interval=retry_interval,
-            workflow_context=workflow_context)
+            send_task_events=send_task_events)
         self.task = task
-        self.cloudify_context = cloudify_context
+        self._cloudify_context = cloudify_context
 
     def apply_async(self):
         """
@@ -258,32 +308,36 @@ class RemoteWorkflowTask(WorkflowTask):
         :return: a RemoteWorkflowTaskResult instance wrapping the
                  celery async result
         """
+        try:
+            self._verify_task_registered()
+            self.workflow_context.internal.send_task_event(TASK_SENDING, self)
+            self.set_state(TASK_SENT)
+            async_result = self.task.apply_async(task_id=self.id)
+            self.async_result = RemoteWorkflowTaskResult(self, async_result)
+        except exceptions.NonRecoverableError as e:
+            self.set_state(TASK_FAILED)
+            self.workflow_context.internal\
+                .send_task_event(TASK_FAILED, self, {'exception': e})
+            self.error = e
+            self.async_result = RemoteWorkflowNotExistTaskResult(self)
 
-        self._verify_task_registered()
-
-        # here to avoid cyclic dependencies
-        from events import send_task_event
-        send_task_event(TASK_SENDING, self)
-
-        async_result = self.task.apply_async(task_id=self.id)
-        self.set_state(TASK_SENT)
-        self.async_result = RemoteWorkflowTaskResult(self, async_result)
         return self.async_result
 
     def is_local(self):
         return False
 
-    def duplicate(self):
+    def _duplicate(self):
         dup = RemoteWorkflowTask(task=self.task,
                                  cloudify_context=self.cloudify_context,
+                                 workflow_context=self.workflow_context,
+                                 task_id=None,  # we want a new task id
                                  info=self.info,
                                  on_success=self.on_success,
                                  on_failure=self.on_failure,
                                  total_retries=self.total_retries,
                                  retry_interval=self.retry_interval,
-                                 workflow_context=self.workflow_context)
+                                 send_task_events=self.send_task_events)
         dup.cloudify_context['task_id'] = dup.id
-        dup.current_retries = self.current_retries
         return dup
 
     @property
@@ -291,26 +345,25 @@ class RemoteWorkflowTask(WorkflowTask):
         """The task name"""
         return self.cloudify_context['task_name']
 
+    @property
+    def cloudify_context(self):
+        return self._cloudify_context
+
     @property
     def target(self):
         """The task target (queue name)"""
         return self.cloudify_context['task_target']
 
     def _verify_task_registered(self):
-        cache = RemoteWorkflowTask.cache
-        registered = cache.get(self.target, set())
-        if self.name not in registered:
-            registered = self._get_registered()
-            cache[self.target] = registered
-
-        if self.name not in registered:
-            raise RuntimeError('Missing task: {} in worker celery.{} \n'
-                               'Registered tasks are: {}'
-                               .format(self.name, self.target, registered))
+        verify_task_registered(self.name, self.target, self._get_registered)
 
     def _get_registered(self):
-        worker_name = 'celery.{}'.format(self.target)
-        inspect = celery_client.control.inspect(destination=[worker_name])
+        # import here because this only applies in remote execution
+        # environments
+        from cloudify.celery import celery
+
+        worker_name = 'celery@{0}'.format(self.target)
+        inspect = celery.control.inspect(destination=[worker_name])
         registered = inspect.registered() or {}
         result = registered.get(worker_name, set())
         return set(result)
@@ -328,8 +381,10 @@ class LocalWorkflowTask(WorkflowTask):
                  on_failure=retry_failure_handler,
                  total_retries=DEFAULT_TOTAL_RETRIES,
                  retry_interval=DEFAULT_RETRY_INTERVAL,
+                 send_task_events=DEFAULT_SEND_TASK_EVENTS,
                  kwargs=None,
-                 task_id=None):
+                 task_id=None,
+                 name=None):
         """
         :param local_task: A callable
         :param workflow_context: the CloudifyWorkflowContext instance
@@ -353,6 +408,7 @@ class LocalWorkflowTask(WorkflowTask):
                               the handlers return a retry attempt.
         :param retry_interval: Number of seconds to wait between retries
         :param kwargs: Local task keyword arguments
+        :param name: optional parameter (default: local_task.__name__)
         """
         super(LocalWorkflowTask, self).__init__(
             info=info,
@@ -361,37 +417,56 @@ class LocalWorkflowTask(WorkflowTask):
             total_retries=total_retries,
             retry_interval=retry_interval,
             task_id=task_id,
-            workflow_context=workflow_context)
+            workflow_context=workflow_context,
+            send_task_events=send_task_events)
         self.local_task = local_task
         self.node = node
-        self.kwargs = kwargs
+        self.kwargs = kwargs or {}
+        self._name = name or local_task.__name__
+
+    def dump(self):
+        super_dump = super(LocalWorkflowTask, self).dump()
+        super_dump.update({
+            'name': self._name
+        })
+        return super_dump
 
     def apply_async(self):
         """
-        Execute the task in the current thread
+        Execute the task in the local task thread pool
         :return: A wrapper for the task result
         """
 
-        self.set_state(TASK_SENT)
-        try:
-            if self.kwargs:
+        def local_task_wrapper():
+            try:
+                self.workflow_context.internal.send_task_event(TASK_STARTED,
+                                                               self)
                 result = self.local_task(**self.kwargs)
-            else:
-                result = self.local_task()
-            self.set_state(TASK_SUCCEEDED)
-            self.async_result = LocalWorkflowTaskResult(self, result)
-        except:
-            exc_type, exception, tb = sys.exc_info()
-            self.set_state(TASK_FAILED)
-            self.async_result = LocalWorkflowTaskResult(self, result=None,
-                                                        error=(exception, tb))
+                self.workflow_context.internal.send_task_event(
+                    TASK_SUCCEEDED, self, event={'result': str(result)})
+                self.async_result._holder.result = result
+                self.set_state(TASK_SUCCEEDED)
+            except BaseException as e:
+                new_task_state = TASK_RESCHEDULED if isinstance(
+                    e, exceptions.OperationRetry) else TASK_FAILED
+                exc_type, exception, tb = sys.exc_info()
+                self.workflow_context.internal.send_task_event(
+                    new_task_state, self, event={'exception': str(exception)})
+                self.async_result._holder.error = (exception, tb)
+                self.set_state(new_task_state)
+
+        self.async_result = LocalWorkflowTaskResult(self)
+
+        self.workflow_context.internal.send_task_event(TASK_SENDING, self)
+        self.set_state(TASK_SENT)
+        self.workflow_context.internal.add_local_task(local_task_wrapper)
 
         return self.async_result
 
     def is_local(self):
         return True
 
-    def duplicate(self):
+    def _duplicate(self):
         dup = LocalWorkflowTask(local_task=self.local_task,
                                 workflow_context=self.workflow_context,
                                 node=self.node,
@@ -399,27 +474,38 @@ class LocalWorkflowTask(WorkflowTask):
                                 on_success=self.on_success,
                                 on_failure=self.on_failure,
                                 total_retries=self.total_retries,
-                                retry_interval=self.retry_interval)
-        dup.current_retries = self.current_retries
+                                retry_interval=self.retry_interval,
+                                send_task_events=self.send_task_events,
+                                kwargs=self.kwargs,
+                                name=self.name)
         return dup
 
     @property
     def name(self):
         """The task name"""
-        return self.local_task.__name__
+        return self._name
+
+    @property
+    def cloudify_context(self):
+        return self.kwargs.get('__cloudify_context')
 
 
 # NOP tasks class
 class NOPLocalWorkflowTask(LocalWorkflowTask):
 
-    def __init__(self):
-        super(NOPLocalWorkflowTask, self).__init__(lambda: None, None)
+    def __init__(self, workflow_context):
+        super(NOPLocalWorkflowTask, self).__init__(lambda: None,
+                                                   workflow_context)
 
     @property
     def name(self):
         """The task name"""
         return 'NOP'
 
+    def apply_async(self):
+        self.set_state(TASK_SUCCEEDED)
+        return LocalWorkflowTaskResult(self)
+
     def is_nop(self):
         return True
 
@@ -490,6 +576,19 @@ class WorkflowTaskResult(object):
         raise NotImplementedError('Implemented by subclasses')
 
 
+class RemoteWorkflowNotExistTaskResult(WorkflowTaskResult):
+    def __init__(self, task):
+        super(RemoteWorkflowNotExistTaskResult, self).__init__(task)
+        self.task = task
+
+    def _get(self):
+        raise self.task.error
+
+    @property
+    def result(self):
+        return self.task.error
+
+
 class RemoteWorkflowTaskResult(WorkflowTaskResult):
     """A wrapper for celery's AsyncResult"""
 
@@ -503,29 +602,42 @@ class RemoteWorkflowTaskResult(WorkflowTaskResult):
     def _refresh_state(self):
         self.async_result = self.task.async_result.async_result
 
+    @property
+    def result(self):
+        return self.async_result.result
+
 
 class LocalWorkflowTaskResult(WorkflowTaskResult):
     """A wrapper for local workflow task results"""
 
-    def __init__(self, task, result, error=None):
+    class ResultHolder(object):
+
+        def __init__(self, result=None, error=None):
+            self.result = result
+            self.error = error
+
+    def __init__(self, task):
         """
         :param task: The LocalWorkflowTask instance
-        :param result: The result if the task finished successfully
-        :param error: a tuple (exception, traceback) if the task failed
         """
         super(LocalWorkflowTaskResult, self).__init__(task)
-        self.result = result
-        self.error = error
+        self._holder = self.ResultHolder()
 
     def _get(self):
-        if self.error is not None:
-            exception, traceback = self.error
+        if self._holder.error is not None:
+            exception, traceback = self._holder.error
             raise exception, None, traceback
-        return self.result
+        return self._holder.result
 
     def _refresh_state(self):
-        self.result = self.task.async_result.result
-        self.error = self.task.async_result.error
+        self._holder = self.task.async_result._holder
+
+    @property
+    def result(self):
+        if self._holder.error:
+            return self._holder.error[0]
+        else:
+            return self._holder.result
 
 
 class HandlerResult(object):
@@ -564,3 +676,18 @@ class HandlerResult(object):
     @classmethod
     def ignore(cls):
         return HandlerResult(cls.HANDLER_IGNORE)
+
+
+def verify_task_registered(name, target, get_registered):
+
+    cache = RemoteWorkflowTask.cache
+    registered = cache.get(target, set())
+    if name not in registered:
+        registered = get_registered()
+        cache[target] = registered
+
+    if name not in registered:
+        raise exceptions.NonRecoverableError(
+            'Missing task: {0} in worker celery.{1} \n'
+            'Registered tasks are: {2}'
+            .format(name, target, registered))
