@@ -13,8 +13,9 @@
 #    * See the License for the specific language governing permissions and
 #    * limitations under the License.
 
-__author__ = 'dank'
 
+import os
+import json
 import time
 
 import networkx as nx
@@ -30,8 +31,6 @@ class TaskDependencyGraph(object):
     :param workflow_context: A WorkflowContext instance (used for logging)
     """
 
-    done_states = [tasks.TASK_FAILED, tasks.TASK_SUCCEEDED]
-
     def __init__(self, workflow_context):
         self.ctx = workflow_context
         self.graph = nx.DiGraph()
@@ -99,17 +98,23 @@ class TaskDependencyGraph(object):
             3. an unhandled exception is raised
             4. the execution is cancelled
 
-        Note: This method will return None unless the execution has been
-        cancelled, in which case the return value will be
-        ``api.EXECUTION_CANCELLED_RESULT``. Callers of this method should
-        check the return value and propagate the result in the latter case.
+        Note: This method will raise an api.ExecutionCancelled error if the
+        execution has been cancelled. When catching errors raised from this
+        method, make sure to re-raise the error if it's
+        api.ExecutionsCancelled in order to allow the execution to be set in
+        cancelled mode properly.
 
         Also note that for the time being, if such a cancelling event
         occurs, the method might return even while there's some operations
         still being executed.
         """
 
-        while not self._is_execution_cancelled():
+        while True:
+
+            if self._is_execution_cancelled():
+                raise api.ExecutionCancelled()
+
+            self._check_dump_request()
 
             # handle all terminated tasks
             # it is important this happens before handling
@@ -130,8 +135,6 @@ class TaskDependencyGraph(object):
             else:
                 time.sleep(0.1)
 
-        return api.EXECUTION_CANCELLED_RESULT
-
     def _is_execution_cancelled(self):
         return api.has_cancel_request()
 
@@ -146,9 +149,9 @@ class TaskDependencyGraph(object):
         """
         now = time.time()
         return (task for task in self.tasks_iter()
-                if task.get_state() == tasks.TASK_PENDING
-                and task.execute_after <= now
-                and not self._task_has_dependencies(task.id))
+                if task.get_state() == tasks.TASK_PENDING and
+                task.execute_after <= now and
+                not self._task_has_dependencies(task.id))
 
     def _terminated_tasks(self):
         """
@@ -157,7 +160,7 @@ class TaskDependencyGraph(object):
         :return: An iterator for terminated tasks
         """
         return (task for task in self.tasks_iter()
-                if task.get_state() in self.done_states)
+                if task.get_state() in tasks.TERMINATED_STATES)
 
     def _task_has_dependencies(self, task_id):
         """
@@ -199,6 +202,17 @@ class TaskDependencyGraph(object):
                            for dependent in dependents]
             self.graph.add_edges_from(added_edges)
 
+    def _check_dump_request(self):
+        task_dump = os.environ.get('WORKFLOW_TASK_DUMP')
+        if not (task_dump and os.path.exists(task_dump)):
+            return
+        os.remove(task_dump)
+        task_dump_path = '{0}.{1}'.format(task_dump, time.time())
+        with open(task_dump_path, 'w') as f:
+            f.write(json.dumps({
+                'tasks': [task.dump() for task in self.tasks_iter()],
+                'edges': [[s, t] for s, t in self.graph.edges_iter()]}))
+
 
 class forkjoin(object):
     """
