@@ -13,13 +13,13 @@
 #    * See the License for the specific language governing permissions and
 #    * limitations under the License.
 
-__author__ = 'ran'
+
+import Queue
 
 EXECUTION_CANCELLED_RESULT = 'execution_cancelled'
 
 
-ctx = None
-pipe = None
+queue = None
 
 
 def has_cancel_request():
@@ -31,16 +31,25 @@ def has_cancel_request():
     request for the execution will have no effect - 'force-cancel' will have
     to be used to abruptly terminate the execution instead.
 
-    Note: When using this method, the workflow must return
-    EXECUTION_CANCELLED_RESULT if indeed the execution gets cancelled.
+    Note: When this method returns True, the workflow should make the
+    appropriate cleanups and then it must raise an ExecutionCancelled error
+    if the execution indeed gets cancelled (i.e. if it's too late to cancel
+    there is no need to raise this exception and the workflow should end
+    normally).
 
     :return: whether there was a request to cancel the workflow execution
     """
-    if pipe.poll():
-        data = pipe.recv()
-        return data['action'] == 'cancel'
-    return False
+    if not queue:
+        return False
+    try:
+        return queue.get_nowait()['action'] == 'cancel'
+    except Queue.Empty:
+        return False
 
 
 class ExecutionCancelled(Exception):
+    """
+    This exception should be raised when a workflow has been cancelled,
+    once appropriate cleanups have taken place.
+    """
     pass
